import os
import platform
import re
import struct
import subprocess

is_windows = platform.system() == "Windows"

is_32_bit = struct.calcsize("P") == 4

compiler_environ = os.environ.copy()

gfortran_min = "6"
gfortran_bits = ("x86_64", "i686")[is_32_bit]


def call_command(cmd):
    try:
        return subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT, env=compiler_environ).decode().strip()
    except Exception:
        return None


def get_ifort_version():
    return call_command("ifort -v")


def get_gfortran_version(command="gfortran"):
    ver = call_command(command + " -dumpversion")
    if ver and "." not in ver:
        ver = call_command(command + " -dumpfullversion")
    return ver


def check_ifort():
    return get_ifort_version() or False


def check_gfortran(version=gfortran_min, msg=False, retry=False):
    from packaging.version import parse as parse_version

    gfortran_version = get_gfortran_version()
    version = str(version)
    if gfortran_version:
        ok = parse_version(version) <= parse_version(gfortran_version)
    else:
        ok = False
    if not ok and is_windows and not retry:
        newpath = None
        best_version = gfortran_min
        for root in (os.path.expanduser("~"), os.environ["ProgramFiles"]):
            mingw = os.path.join(root, "mingw-w64")
            if not os.path.isdir(mingw):
                mingw = os.path.join(root, "mingw64")

            if os.path.isdir(mingw):
                # look for mingw installation
                dirs = [
                    name
                    for name in os.listdir(mingw)
                    if gfortran_bits in name and os.path.isdir(os.path.join(mingw, name))
                ]
                for i, x in enumerate(dirs):
                    if "." in x:
                        ver = x.split("-")[1]
                        if parse_version(best_version) <= parse_version(ver):
                            best_version = ver
                            newpath = os.path.join(mingw, x)
                            if os.path.exists(os.path.join(newpath, "mingw64", "bin")):
                                newpath = os.path.join(newpath, "mingw64", "bin")
                            elif os.path.exists(os.path.join(newpath, "bin")):
                                newpath = os.path.join(newpath, "bin")
                bin_path = os.path.join(mingw, "bin")
                if os.path.exists(bin_path):
                    ver = get_gfortran_version('"' + os.path.join(bin_path, "gfortran") + '"')
                    if ver and parse_version(best_version) <= parse_version(ver):
                        best_version = ver
                        newpath = bin_path
        if newpath:
            if os.path.exists(os.path.join(newpath, "gfortran.exe")):
                if not compiler_environ["PATH"].startswith(newpath):
                    compiler_environ["PATH"] = newpath + ";" + compiler_environ["PATH"]
                return check_gfortran(version, msg, retry=True)
    if ok and is_windows:
        version_str = str(subprocess.check_output("gfortran -dumpmachine", shell=True, env=compiler_environ))
        ok = gfortran_bits in version_str
    if not ok and msg:
        raise Exception(
            "You need ifort or gfortran %s or higher to compile (found: %s).\nSee %s"
            % (version, gfortran_version, "https://camb.readthedocs.io/en/latest/fortran_compilers.html")
        )

    return ok, gfortran_version


def makefile_dict(filename):
    # this is very non-general, just enough for pulling source file names from Makefile
    with open(filename, "r") as f:
        lines = f.readlines()
    vals = {}
    lastval = None
    append = False
    for line in lines:
        parts = line.split("\\")
        line = parts[0].strip()
        if "?=" in line:
            key, val = line.split("?=")
            env = os.environ.get(key.strip(), None)
            if env:
                vals[key] = env
                lastval = None
                append = False
                continue
            else:
                line = line.replace("?=", "=")
        if append and lastval:
            vals[lastval] += " " + line
        elif "=" in line:
            if len(line.split("=")) == 2 and ":" not in line:
                lastval, value = line.split("=")
                lastval = lastval.strip()
                vals[lastval] = value.strip()
        else:
            lastval = None
        append = len(parts) > 1

    def repl(groups):
        if groups.group(1) in vals:
            return vals[groups.group(1)]
        else:
            return groups.group(0)

    for key, value in vals.items():
        if "$" in value:
            vals[key] = re.sub(r"\$\((\w+)\)", repl, value)
    return vals



"""

CAMB, Code for Anisotropies in the Microwave Background (https://camb.info)
Computational modules are wrapped Fortran 2003, but can be used entirely from Python.

"""

__author__ = "Antony Lewis"
__contact__ = "antony at cosmologist dot info"
__url__ = "https://camb.readthedocs.io"
__version__ = "1.6.1"

from . import baseconfig

baseconfig.check_fortran_version(__version__)
from . import dark_energy, initialpower, model, nonlinear, reionization
from ._config import config
from .baseconfig import CAMBError, CAMBFortranError, CAMBParamRangeError, CAMBUnknownArgumentError, CAMBValueError
from .camb import (
    free_global_memory,
    get_age,
    get_background,
    get_matter_power_interpolator,
    get_results,
    get_transfer_functions,
    get_valid_numerical_params,
    get_zre_from_tau,
    read_ini,
    run_ini,
    set_feedback_level,
    set_params,
    set_params_cosmomc,
)
from .dark_energy import DarkEnergyFluid, DarkEnergyPPF
from .initialpower import InitialPowerLaw, SplinedInitialPower
from .mathutils import threej
from .model import CAMBparams, TransferParams
from .nonlinear import Halofit
from .reionization import ExpReionization, TanhReionization
from .results import CAMBdata, ClTransferData, MatterTransferData



from ctypes import POINTER, byref, c_bool, c_double, c_void_p

from .baseconfig import F2003Class, f_pointer, fortran_class


class ReionizationModel(F2003Class):
    """
    Abstract base class for reionization models.
    """

    _fields_ = [
        ("Reionization", c_bool, "Is there reionization? (can be off for matter power, which is independent of it)")
    ]


class BaseTauWithHeReionization(ReionizationModel):
    """
    Abstract class for models that map z_re to tau, and include second reionization of Helium
    """

    _fields_ = [
        ("use_optical_depth", c_bool, "Whether to use the optical depth or redshift parameters"),
        ("redshift", c_double, "Reionization redshift (xe=0.5) if use_optical_depth=False"),
        ("optical_depth", c_double, "Optical depth if use_optical_depth=True"),
        (
            "fraction",
            c_double,
            "Reionization fraction when complete, "
            "or -1 for full ionization of hydrogen and first ionization of helium.",
        ),
        ("include_helium_fullreion", c_bool, "Whether to include second reionization of helium"),
        ("helium_redshift", c_double, "Redshift for second reionization of helium"),
        ("helium_delta_redshift", c_double, "Width in redshift for second reionization of helium"),
        ("helium_redshiftstart", c_double, "Include second helium reionization below this redshift"),
        ("tau_solve_accuracy_boost", c_double, "Accuracy boosting parameter for solving for z_re from tau"),
        (
            "timestep_boost",
            c_double,
            "Accuracy boosting parameter for the minimum number of time sampling steps through reionization",
        ),
        ("max_redshift", c_double, "Maximum redshift allowed when mapping tau into reionization redshift"),
        ("__min_redshift", c_double, "Minimum redshift allowed when mapping tau into reionization redshift"),
        ("__fHe", c_double, "Helium fraction"),
        ("__state", f_pointer),
    ]

    _fortran_class_module_ = "Reionization"
    _fortran_class_name_ = "TBaseTauWithHeReionization"

    _methods_ = [("GetZreFromTau", [c_void_p, POINTER(c_double)], c_double, {"nopass": True})]

    def get_zre(self, params, tau=None):
        """
        Get the midpoint redshift of reionization.

        :param params: :class:`.model.CAMBparams` instance with cosmological parameters
        :param tau: if set, calculate the redshift for optical depth tau, otherwise uses currently set parameters
        :return: reionization mid-point redshift
        """
        if self.use_optical_depth or tau:
            from .camb import CAMBparams

            assert isinstance(params, CAMBparams)
            return self.f_GetZreFromTau(byref(params), c_double(tau or self.optical_depth))
        else:
            return self.redshift

    def set_zrei(self, zrei):
        """
        Set the mid-point reionization redshift

        :param zrei: mid-point redshift
        :return: self
        """
        self.use_optical_depth = False
        self.redshift = zrei
        return self

    def set_tau(self, tau):
        """
        Set the optical depth

        :param tau: optical depth
        :return: self
        """
        self.use_optical_depth = True
        self.optical_depth = tau
        return self

    def set_extra_params(self, max_zrei=None):
        """
        Set extra parameters (not tau, or zrei)

        :param max_zrei: maximum redshift allowed when mapping tau into reionization redshift
        """
        if max_zrei is not None:
            self.max_redshift = max_zrei


@fortran_class
class TanhReionization(BaseTauWithHeReionization):
    """
    This default (unphysical) tanh x_e parameterization is described in
    Appendix B of `arXiv:0804.3865 <https://arxiv.org/abs/0804.3865>`_
    """

    _fields_ = [("delta_redshift", c_double, "Duration of reionization")]

    _fortran_class_name_ = "TTanhReionization"

    def set_extra_params(self, deltazrei=None, max_zrei=None) -> None:
        """
        Set extra parameters (not tau, or zrei)

        :param deltazrei: delta z for reionization
        :param max_zrei: maximum redshift allowed when mapping tau into reionization
        """
        super().set_extra_params(max_zrei)
        if deltazrei is not None:
            self.delta_redshift = deltazrei


@fortran_class
class ExpReionization(BaseTauWithHeReionization):
    """
    An ionization fraction that decreases exponentially at high z, saturating to fully ionized at fixed redshift.
    This model has a minimum non-zero tau around 0.04 for reion_redshift_complete=6.1.
    Similar to e.g. arXiv:1509.02785, arXiv:2006.16828, but not attempting to fit shape near x_e~1 at z<6.1
    """

    _fields_ = [
        ("reion_redshift_complete", c_double, "end of reionization"),
        ("reion_exp_smooth_width", c_double, "redshift scale to smooth exponential"),
        ("reion_exp_power", c_double, "power in exponential, exp(-lambda(z-redshift_complete)^exp_power)"),
    ]

    _fortran_class_name_ = "TExpReionization"

    def set_extra_params(
        self, reion_redshift_complete=None, reion_exp_power=None, reion_exp_smooth_width=None, max_zrei=None
    ) -> None:
        """
        Set extra parameters (not tau, or zrei)

        :param reion_redshift_complete: redshift at which reionization complete (e.g. around 6)
        :param reion_exp_power: power in exponential decay with redshift
        :param reion_exp_smooth_width: smoothing parameter to keep derivative smooth
        :param max_zrei: maximum redshift allowed when mapping tau into reionization
        """
        super().set_extra_params(max_zrei)
        if reion_redshift_complete is not None:
            self.reion_redshift_complete = reion_redshift_complete
        if reion_exp_power is not None:
            self.reion_exp_power = reion_exp_power
        if reion_exp_smooth_width is not None:
            self.reion_exp_smooth_width = reion_exp_smooth_width



import argparse
import ctypes
import os
import sys
from argparse import RawTextHelpFormatter

sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))
import camb
from camb.baseconfig import lib_import


def run_command_line():
    parser = argparse.ArgumentParser(
        formatter_class=RawTextHelpFormatter,
        description="Python command line CAMB reading parameters from a .ini file."
        + "\n\nSample .ini files are provided in the source distribution, "
        "e.g. see inifiles/planck_2018.ini at "
        "https://github.com/cmbant/CAMB/tree/master/inifiles",
    )
    parser.add_argument("ini_file", help="text .ini file with parameter settings")
    parser.add_argument(
        "--validate", action="store_true", help="Just validate the .ini file, dont actually run anything"
    )
    parser.add_argument("-V", "--version", action="version", version=camb.__version__)
    args = parser.parse_args()

    if not os.path.exists(args.ini_file):
        sys.exit("File not found: %s" % args.ini_file)

    s = ctypes.create_string_buffer(args.ini_file.encode("latin-1"))

    # Import wrapper function round fortran command line program
    CAMB_RunCommandLine = lib_import("camb", "camb", "CommandLineValidate" if args.validate else "CommandLineRun")
    CAMB_RunCommandLine.argtypes = [ctypes.c_char_p, ctypes.c_long]
    CAMB_RunCommandLine(s, ctypes.c_long(len(args.ini_file)))
    if args.validate:
        print("OK")


if __name__ == "__main__":
    run_command_line()



from ctypes import POINTER, byref, c_bool, c_double, c_int

from .baseconfig import AllocatableArrayDouble, CAMBError, F2003Class, f_pointer, fortran_class, np, numpy_1d


class DarkEnergyModel(F2003Class):
    """
    Abstract base class for dark energy model implementations.
    """

    _fields_ = [("__is_cosmological_constant", c_bool), ("__num_perturb_equations", c_int)]

    def validate_params(self) -> None:
        pass


class DarkEnergyEqnOfState(DarkEnergyModel):
    """
    Abstract base class for models using w and wa parameterization with use w(a) = w + (1-a)*wa parameterization,
    or call set_w_a_table to set another tabulated w(a). If tabulated w(a) is used, w and wa are set
    to approximate values at z=0.

    See :meth:`.model.CAMBparams.set_initial_power_function` for a convenience constructor function to
    set a general interpolated P(k) model from a python function.

    """

    _fortran_class_module_ = "DarkEnergyInterface"
    _fortran_class_name_ = "TDarkEnergyEqnOfState"

    _fields_ = [
        ("w", c_double, "w(0)"),
        ("wa", c_double, "-dw/da(0)"),
        ("cs2", c_double, "fluid rest-frame sound speed squared"),
        ("use_tabulated_w", c_bool, "using an interpolated tabulated w(a) rather than w, wa above"),
        ("__no_perturbations", c_bool, "turn off perturbations (unphysical, so hidden in Python)"),
    ]

    _methods_ = [("SetWTable", [numpy_1d, numpy_1d, POINTER(c_int)])]

    def set_params(self, w=-1.0, wa=0, cs2=1.0):
        """
         Set the parameters so that P(a)/rho(a) = w(a) = w + (1-a)*wa

        :param w: w(0)
        :param wa: -dw/da(0)
        :param cs2: fluid rest-frame sound speed squared
        """
        self.w = w
        self.wa = wa
        self.cs2 = cs2
        self.validate_params()

    def validate_params(self):
        if not self.use_tabulated_w and self.wa + self.w > 0:
            raise CAMBError("dark energy model has w + wa > 0, giving w>0 at high redshift")

    def set_w_a_table(self, a, w) -> "DarkEnergyEqnOfState":
        """
        Set w(a) from numerical values (used as cubic spline). Note this is quite slow.

        :param a: array of scale factors
        :param w: array of w(a)
        :return: self
        """
        if len(a) != len(w):
            raise ValueError("Dark energy w(a) table non-equal sized arrays")
        if not np.isclose(a[-1], 1):
            raise ValueError("Dark energy w(a) arrays must end at a=1")
        if np.any(a <= 0):
            raise ValueError("Dark energy w(a) table cannot be set for a<=0")

        a = np.ascontiguousarray(a, dtype=np.float64)
        w = np.ascontiguousarray(w, dtype=np.float64)

        self.f_SetWTable(a, w, byref(c_int(len(a))))
        return self

    def __getstate__(self):
        if self.use_tabulated_w:
            raise TypeError("Cannot save class with splines")
        return super().__getstate__()


@fortran_class
class DarkEnergyFluid(DarkEnergyEqnOfState):
    """
    Class implementing the w, wa or splined w(a) parameterization using the constant sound-speed single fluid model
    (as for single-field quintessence).

    """

    _fortran_class_module_ = "DarkEnergyFluid"
    _fortran_class_name_ = "TDarkEnergyFluid"

    def validate_params(self) -> None:
        super().validate_params()
        if not self.use_tabulated_w:
            if self.wa and (self.w < -1 - 1e-6 or 1 + self.w + self.wa < -1e-6):
                raise CAMBError("fluid dark energy model does not support w crossing -1")

    def set_w_a_table(self, a, w) -> "DarkEnergyEqnOfState":
        # check w array has elements that do not cross -1
        if np.sign(1 + np.max(w)) - np.sign(1 + np.min(w)) == 2:
            raise CAMBError("fluid dark energy model does not support w crossing -1")
        return super().set_w_a_table(a, w)


@fortran_class
class DarkEnergyPPF(DarkEnergyEqnOfState):
    """
    Class implementing the w, wa or splined w(a) parameterization in the PPF perturbation approximation
    (`arXiv:0808.3125 <https://arxiv.org/abs/0808.3125>`_)
    Use inherited methods to set parameters or interpolation table.

    Note PPF is not a physical model and just designed to allow crossing -1 in an ad hoc smooth way. For models
    with w>-1 but far from cosmological constant, it can give quite different answers to the fluid model with c_s^2=1.

    """

    # cannot declare c_Gamma_ppf directly here as have not defined all fields in DarkEnergyEqnOfState (TCubicSpline)
    _fortran_class_module_ = "DarkEnergyPPF"
    _fortran_class_name_ = "TDarkEnergyPPF"


@fortran_class
class AxionEffectiveFluid(DarkEnergyModel):
    """
    Example implementation of a specific (early) dark energy fluid model
    (`arXiv:1806.10608 <https://arxiv.org/abs/1806.10608>`_).
    Not well tested, but should serve to demonstrate how to make your own custom classes.
    """

    _fields_ = [
        ("w_n", c_double, "effective equation of state parameter"),
        ("fde_zc", c_double, "energy density fraction at z=zc"),
        ("zc", c_double, "decay transition redshift (not same as peak of energy density fraction)"),
        ("theta_i", c_double, "initial condition field value"),
    ]

    _fortran_class_name_ = "TAxionEffectiveFluid"
    _fortran_class_module_ = "DarkEnergyFluid"

    def set_params(self, w_n, fde_zc, zc, theta_i=None):
        self.w_n = w_n
        self.fde_zc = fde_zc
        self.zc = zc
        if theta_i is not None:
            self.theta_i = theta_i


# base class for scalar field quintessence models
class Quintessence(DarkEnergyModel):
    r"""
    Abstract base class for single scalar field quintessence models.

    For each model the field value and derivative are stored and splined at sampled scale factor values.

    To implement a new model, need to define a new derived class in Fortran,
    defining Vofphi and setting up initial conditions and interpolation tables (see TEarlyQuintessence as example).

    """

    _fields_ = [
        ("DebugLevel", c_int),
        ("astart", c_double),
        ("integrate_tol", c_double),
        ("sampled_a", AllocatableArrayDouble),
        ("phi_a", AllocatableArrayDouble),
        ("phidot_a", AllocatableArrayDouble),
        ("__npoints_linear", c_int),
        ("__npoints_log", c_int),
        ("__dloga", c_double),
        ("__da", c_double),
        ("__log_astart", c_double),
        ("__max_a_log", c_double),
        ("__ddphi_a", AllocatableArrayDouble),
        ("__ddphidot_a", AllocatableArrayDouble),
        ("__state", f_pointer),
    ]
    _fortran_class_module_ = "Quintessence"

    def __getstate__(self):
        raise TypeError("Cannot save class with splines")


@fortran_class
class EarlyQuintessence(Quintessence):
    r"""
    Example early quintessence (axion-like, as `arXiv:1908.06995 <https://arxiv.org/abs/1908.06995>`_) with potential

     V(\phi) = m^2f^2 (1 - cos(\phi/f))^n + \Lambda_{cosmological constant}

    """

    _fields_ = [
        ("n", c_double, "power index for potential"),
        ("f", c_double, r"f/Mpl (sqrt(8\piG)f); only used for initial search value when use_zc is True"),
        (
            "m",
            c_double,
            "mass parameter in reduced Planck mass units; only used for initial search value when use_zc is True",
        ),
        ("theta_i", c_double, "phi/f initial field value"),
        ("frac_lambda0", c_double, "fraction of dark energy in cosmological constant today (approximated as 1)"),
        ("use_zc", c_bool, "solve for f, m to get specific critical redshift zc and fde_zc"),
        ("zc", c_double, "redshift of peak fractional early dark energy density"),
        ("fde_zc", c_double, "fraction of early dark energy density to total at peak"),
        ("npoints", c_int, "number of points for background integration spacing"),
        ("min_steps_per_osc", c_int, "minimum number of steps per background oscillation scale"),
        (
            "fde",
            AllocatableArrayDouble,
            "after initialized, the calculated background early dark energy fractions at sampled_a",
        ),
        ("__ddfde", AllocatableArrayDouble),
    ]
    _fortran_class_name_ = "TEarlyQuintessence"

    def set_params(self, n, f=0.05, m=5e-54, theta_i=0.0, use_zc=True, zc=None, fde_zc=None):
        self.n = n
        self.f = f
        self.m = m
        self.theta_i = theta_i
        self.use_zc = use_zc
        if use_zc:
            if zc is None or fde_zc is None:
                raise ValueError("must set zc and fde_zc if using 'use_zc'")
            self.zc = zc
            self.fde_zc = fde_zc


# short names for models that support w/wa
F2003Class._class_names.update({"fluid": DarkEnergyFluid, "ppf": DarkEnergyPPF})



import os
import shutil
import subprocess
import sys
from typing import Any

from setuptools import Command, Extension, setup
from setuptools.command.build_ext import build_ext
from setuptools.command.build_py import build_py
from setuptools.command.develop import develop
from setuptools.command.install import install
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

file_dir = os.path.abspath(os.path.dirname(__file__))
os.chdir(file_dir)

sys.path.insert(0, os.path.join(file_dir, "camb"))
_compile: Any = __import__("_compilers")

if _compile.is_windows:
    DLLNAME = "cambdll.dll"
else:
    DLLNAME = "camblib.so"


def get_forutils():
    fpath = os.getenv("FORUTILSPATH")

    def git_install_forutils():
        try:
            print("forutils not found, attempting to install using git...")
            os.chdir("..")
            fbranch = os.getenv("FORUTILSBRANCH", "1.0.3" if os.environ.get("CONDA_BUILD") else "master")
            try:
                if (
                    subprocess.call(
                        "git clone --branch %s --depth=1 https://github.com/cmbant/forutils" % fbranch, shell=True
                    )
                    == 0
                ):
                    return os.path.join("..", "forutils")
            finally:
                os.chdir("fortran")
        except Exception:
            print("Failed to install using git")

    if not fpath:
        dirs = ["..", ".." + os.sep + ".."]
        for _dir in dirs:
            path = os.path.join(_dir, "forutils")
            if os.path.isdir(path):
                fpath = path
                main_dir = _dir
                break
        if not fpath:
            fpath = git_install_forutils()
        elif not os.path.exists(os.path.join(fpath, "Makefile")):
            if os.path.isdir(os.path.join("..", ".git")):
                # submodule may not be pulled
                try:
                    try:
                        os.chdir("..")
                        print("forutils directory found but no Makefile. Attempting to clone submodule...")
                        if subprocess.call("git submodule update --init --recursive", shell=True, cwd=main_dir) != 0:
                            raise Exception()
                    finally:
                        os.chdir("fortran")
                except Exception:
                    fpath = None
                    print("Failed to install forutils using git")
            else:
                os.rmdir(fpath)
                if not os.path.exists(fpath):
                    fpath = git_install_forutils()
                else:
                    fpath = None

    if not fpath:
        raise Exception(
            "Install forutils from https://github.com/cmbant/forutils, "
            "pull the forutils submodule, or set FORUTILSPATH variable.\n"
            'If you are cloning with git, use "git clone --recursive"'
        )
    return fpath


def clean_dir(path, rmdir=False):
    if os.path.isdir(path):
        for f in os.listdir(path):
            os.remove(os.path.join(path, f))
        if rmdir:
            os.rmdir(path)


def make_library(cluster=False):
    os.chdir(os.path.join(file_dir, "fortran"))
    pycamb_path = ".."
    lib_file = os.path.join(pycamb_path, "camb", DLLNAME)
    if _compile.is_windows or not _compile.check_ifort():
        ok, gfortran_version = _compile.check_gfortran(msg=not _compile.is_windows)
        if ok and "8.2.0" in gfortran_version:
            print("WARNING: gfortran 8.2.0 may be buggy and give unreliable results or crashes, upgrade gfortran.")
    if _compile.is_windows:
        COMPILER = "gfortran"
        FFLAGS = "-shared -static -cpp -fopenmp -O3 -fmax-errors=4"
        # FFLAGS = "-shared -static -cpp -fopenmp -g -fbacktrace -ffpe-trap=invalid,overflow,zero " \
        #         "-fbounds-check -fmax-errors=4"
        if _compile.is_32_bit:
            FFLAGS = "-m32 " + FFLAGS
        if not ok:
            print(
                "WARNING: gfortran %s or higher not in path (if you just installed "
                "you may need to log off and on again)." % _compile.gfortran_min
            )
            print("        You can get a Windows gfortran build from https://sourceforge.net/projects/mingw-w64/files/")
            print("        - go to Files, and download MinGW-W64 Online Installer.")
            print("        Alternatively newer versions at https://github.com/niXman/mingw-builds-binaries")
            if _compile.is_32_bit:
                raise IOError("No 32bit Windows DLL provided, you need to build or use 64 bit python")
            else:
                print("Using pre-compiled binary instead - any local changes will be ignored...")
        else:
            fpath = get_forutils()
            makefile = _compile.makefile_dict("Makefile_main")
            SOURCES = makefile["SOURCEFILES"].split()
            FORUTILS = [
                os.path.join(fpath, f.replace(".f90", ""))
                for f in _compile.makefile_dict(os.path.join(fpath, "Makefile"))["SRCS"]
                .replace("MatrixUtils.f90", "")
                .split()
            ]
            tmpdir = "WinDLL"
            if not os.path.isdir(tmpdir):
                os.mkdir(tmpdir)
            ofiles = []
            new_compiler = True
            ver_file = os.path.join(tmpdir, "compiler.ver")
            if os.path.exists(ver_file):
                with open(ver_file, "r") as f:
                    new_compiler = gfortran_version != f.readline().strip()
            if new_compiler:
                clean_dir(tmpdir)
                with open(ver_file, "w") as f:
                    f.write(gfortran_version)

            need_compile = not os.path.exists(lib_file)
            if not need_compile:
                dll_time = os.path.getmtime(lib_file)
            for source in FORUTILS + SOURCES:
                # manual Make using dependency files if available
                outroot = os.path.join(tmpdir, os.path.split(source)[1])
                fout = outroot + ".o"
                ofiles += [fout]
                modified = new_compiler or not os.path.exists(fout)
                if not modified:
                    o_time = os.path.getmtime(fout)
                    modified = o_time < os.path.getmtime(source + ".f90") or not os.path.exists(outroot + ".d")
                    if not modified:
                        with open(outroot + ".d", "r") as f:
                            for dependence in " ".join(f.readlines()).replace("\\\n", "").split(":")[1].strip().split():
                                if os.path.getmtime(dependence) > o_time:
                                    modified = True
                                    break

                if modified:
                    need_compile = True
                    cmd = COMPILER + " " + FFLAGS + " " + source + ".f90 -MMD -c -o %s -J%s" % (fout, tmpdir)
                    print(cmd)
                    if subprocess.call(cmd, shell=True, env=_compile.compiler_environ) != 0:
                        raise IOError("Compilation failed")
                elif not need_compile and dll_time < o_time:
                    need_compile = True

            if need_compile or not os.path.exists(lib_file):
                if os.path.exists(lib_file):
                    # raise an exception if the file in use and cannot be deleted
                    try:
                        os.remove(lib_file)
                    except OSError:
                        raise IOError("dll file in use. Stop python codes and notebook kernels that are using camb.")
                print("Compiling sources...")
                cmd = COMPILER + " " + FFLAGS + " " + " ".join(ofiles) + " -o %s -J%s" % (lib_file, tmpdir)
                print(cmd)
                if subprocess.call(cmd, shell=True, env=_compile.compiler_environ) != 0:
                    raise IOError("Compilation failed")
            else:
                print("DLL up to date.")
    else:
        if not _compile.call_command("make -v"):
            raise IOError(
                'Build failed - you must have "make" installed. '
                'E.g. on ubuntu install with "sudo apt install make" (or use build-essential package).'
            )
        get_forutils()
        print("Compiling source...")
        subprocess.call(
            "make python PYCAMB_OUTPUT_DIR=%s/camb/ CLUSTER_SAFE=%d"
            % (pycamb_path, int(cluster if not os.getenv("GITHUB_ACTIONS") else 1)),
            shell=True,
        )
        subprocess.call("chmod 755 %s" % lib_file, shell=True)

    if not os.path.isfile(os.path.join(pycamb_path, "camb", DLLNAME)):
        sys.exit("Compilation failed")
    tem_file = "HighLExtrapTemplate_lenspotentialCls.dat"
    tem = os.path.join(pycamb_path, "camb", tem_file)
    if not os.path.exists(tem) or os.path.getmtime(tem) < os.path.getmtime(tem_file):
        shutil.copy(tem_file, tem)

    os.chdir(file_dir)


class MakeLibrary(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        make_library(False)


class MakeLibraryCluster(MakeLibrary):
    def run(self):
        make_library(True)


class SharedLibrary(build_py):
    def run(self):
        make_library(False)
        build_py.run(self)


class SharedLibraryCluster(SharedLibrary):
    def run(self):
        make_library(True)
        build_py.run(self)


class DevelopLibrary(develop):
    def run(self):
        make_library(False)
        develop.run(self)


class DevelopLibraryCluster(develop):
    def run(self):
        make_library(True)
        develop.run(self)


class CleanLibrary(MakeLibrary):
    def run(self):
        if _compile.is_windows:
            clean_dir(os.path.join(file_dir, "fortran", "WinDLL"), rmdir=True)
        else:
            subprocess.call("make clean", shell=True, cwd=os.path.join(file_dir, "fortran"))


class BDistWheelNonPure(_bdist_wheel):
    def finalize_options(self):
        super().finalize_options()
        self.root_is_pure = False

    def get_tag(self):
        _, _, plat = super().get_tag()
        return "py3", "none", plat


class InstallPlatlib(install):
    def finalize_options(self):
        super().finalize_options()
        if self.distribution.has_ext_modules():
            self.install_lib = self.install_platlib


class BuildExtCommand(build_ext):
    """Ensure built extensions are added to the correct path in the wheel."""

    def run(self):
        pass


if __name__ == "__main__":
    setup(
        name=os.getenv("CAMB_PACKAGE_NAME", "camb"),
        zip_safe=False,
        cmdclass={
            "build_py": SharedLibrary,
            "build_cluster": SharedLibraryCluster,
            "make": MakeLibrary,
            "make_cluster": MakeLibraryCluster,
            "clean": CleanLibrary,
            "develop": DevelopLibrary,
            "develop_cluster": DevelopLibraryCluster,
            "bdist_wheel": BDistWheelNonPure,
            "install": InstallPlatlib,
            "build_ext": BuildExtCommand,
        },
        ext_modules=[Extension("camb.camblib", [])],
        packages=["camb", "camb.tests"],
        platforms="any",
        package_data={
            "camb": [
                DLLNAME,
                "HighLExtrapTemplate_lenspotentialCls.dat",
                "PArthENoPE_880.2_marcucci.dat",
                "PArthENoPE_880.2_standard.dat",
                "PRIMAT_Yp_DH_Error.dat",
                "PRIMAT_Yp_DH_ErrorMC_2021.dat",
            ]
        },
    )
